package com.laguerre.kan;

import java.util.ArrayList;
import java.util.Arrays;

import com.laguerre.kan.SceneConductor.ISurfaceChangedListener;
import android.opengl.GLES20;
import android.opengl.Matrix;

public final class Camera extends Component implements ISurfaceChangedListener {
	/**
	 * Contains values that are invariant for a {@link Camera} that renders
	 * objects, such as the view matrix, the projection matrix, the concatenation
	 * of the view and the projection matrix, or the eye/camera position.
	 */
	public static final class RenderingValues {
		private final int[] viewSize = new int[2];
		private final Vector3 eyePosition = Vector3.getNewZeroVec();
		private final float[] viewMatrix = new float[16];
		private final float[] invViewMatrix = new float[16];
		private final float[] projMatrix = new float[16];
		private final float[] invProjMatrix = new float[16];
		private final float[] viewProjMatrix = new float[16];
		private final float[] invViewProjMatrix = new float[16];

		RenderingValues() {
			Arrays.fill(this.viewSize, 100);
			Matrix.setIdentityM(this.viewMatrix, 0);
			Matrix.setIdentityM(this.invViewMatrix, 0);
			Matrix.setIdentityM(this.projMatrix, 0);
			Matrix.setIdentityM(this.invProjMatrix, 0);
			Matrix.setIdentityM(this.viewProjMatrix, 0);
			Matrix.setIdentityM(this.invViewProjMatrix, 0);
		}

		/**
		 * Sets the size of the view in pixels.
		 * @param width
		 * @param height
		 */
		public void setViewSize(int width, int height) {
			this.viewSize[0] = width;
			this.viewSize[1] = height;
		}

		/**
		 * Sets the size of the view in pixels.
		 * @param viewSize where viewSize[0] contains width and viewSize[1]
		 *   contains height.
		 */
		public void setViewSize(int[] viewSize) {
			System.arraycopy(viewSize, 0, this.viewSize, 0, 2);
		}

		/**
		 * Sets width of the view in pixels.
		 * @param width
		 */
		public void setViewSizeWidth(int width) {
			this.viewSize[0] = width;
		}

		/**
		 * Sets height of the view in pixels.
		 * @param width
		 */
		public void setViewSizeHeight(int height) {
			this.viewSize[1] = height;
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains view size in pixels in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the view size.
		 */
		public int[] getSharedViewSize() {
			return this.viewSize;
		}

		/**
		 * Sets the eye/camera position in world space.
		 * @param position
		 */
		public void setEyePosition(Vector3 position) {
			this.eyePosition.setValues(position);
		}

		/**
		 * Gets a direct reference to the inner <code>Vector3</code> that
		 * contains eye position in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the eye position.
		 */
		public Vector3 getSharedEyePosition() {
			return this.eyePosition;
		}

		/**
		 * Sets the view matrix and computes its inverse automatically.
		 * @param viewMatrix
		 */
		public void setViewMatrix(float[] viewMatrix) {
			setViewMatrix(viewMatrix, null);
		}

		/**
		 * Sets the view matrix and its inverse. If the inverse equals to
		 * <code>null</code>, it will be computed automatically.
		 * @param viewMatrix the view matrix.
		 * @param invViewMatrix the inverse of the view matrix or <code>null</code>.
		 */
		public void setViewMatrix(float[] viewMatrix, float[] invViewMatrix) {
			setMatrixes(viewMatrix, this.viewMatrix, invViewMatrix, this.invViewMatrix);
		}

		/**
		 * Sets the view-projection matrix and computes its inverse automatically.
		 * @param projViewMatrix
		 */
		public void setViewProjMatrix(float[] projViewMatrix) {
			setViewProjMatrix(projViewMatrix, null);
		}

		/**
		 * Sets the view-projection matrix and its inverse. If the inverse equals
		 * to <code>null</code>, it will be computed automatically.
		 * @param viewProjMatrix the view-projection matrix.
		 * @param invViewProjMatrix the inverse of the view-projection matrix or <code>null</code>.
		 */
		public void setViewProjMatrix(float[] viewProjMatrix, float[] invViewProjMatrix) {
			setMatrixes(viewProjMatrix, this.viewProjMatrix, invViewProjMatrix, this.invViewProjMatrix);
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the view matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the view matrix.
		 */
		public float[] getSharedViewMatrix() {
			return this.viewMatrix;
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the inverse view matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the inverse view matrix.
		 */
		public float[] getSharedInvViewMatrix() {
			return this.invViewMatrix;
		}

		/**
		 * Sets the projection matrix and computes its inverse automatically.
		 * @param projMatrix
		 */
		public void setProjMatrix(float[] projMatrix) {
			setProjMatrix(projMatrix, null);
		}

		/**
		 * Sets the projection matrix and its inverse. If the inverse equals
		 * to <code>null</code>, it will be computed automatically.
		 * @param projMatrix the projection matrix.
		 * @param invProjMatrix the inverse of the projection matrix or <code>null</code>.
		 */
		public void setProjMatrix(float[] projMatrix, float[] invProjMatrix) {
			setMatrixes(projMatrix, this.projMatrix, invProjMatrix, this.invProjMatrix);
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the projection matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the projection matrix.
		 */
		public float[] getSharedProjMatrix() {
			return this.projMatrix;
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the inverse projection matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the inverse projection matrix.
		 */
		public float[] getSharedInvProjMatrix() {
			return this.invProjMatrix;
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the view-projection matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the view-projection matrix.
		 */
		public float[] getSharedViewProjMatrix() {
			return this.viewProjMatrix;
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the inverse view-projection matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the inverse view-projection matrix.
		 */
		public float[] getSharedInvViewProjMatrix() {
			return this.invViewProjMatrix;
		}

		/**
		 * Fills a <code>float[]</code> array that represents a matrix and also
		 * set its inverse: to the proposed value or automatically if the value
		 * is <code>null</code>.
		 * @param src the matrix to copy
		 * @param fieldDst the reference to the array that will contain the copy of <code>src</code>.
		 * @param invSrc the inverse of the matrix to copy, or <code>null</code> to compute it automatically.
		 * @param fieldInvDst the reference to the array that will contain the copy of the inverse matrix.
		 */
		private void setMatrixes(float[] src, float[] fieldDst, float[] invSrc, float[] fieldInvDst) {
			System.arraycopy(src, 0, fieldDst, 0, src.length);
			if (invSrc != null) {
				System.arraycopy(invSrc, 0, fieldInvDst, 0, src.length);
			} else {
				Matrix.invertM(fieldInvDst, 0, src, 0);
			}
		}
	}

	private static Camera currentCamera = null;
	public static Camera getCurrent() {
		return currentCamera;
	}

	private boolean enabled;
	private RenderTexture renderTexture;
	private final ImageEffects postEffects = new ImageEffects();
	private int fbHandle;
	private int depthFbHandle;
	/**
	 * Used to delayed a mandatory call of <code>updateProj</code> by
	 * <code>preRenderToTexture</code>.
	 */
	private boolean needsUpdateProj = false;
	/**
	 * The default value of the culling mask is <code>1&lt;&lt;0</code>.
	 */
	private int cullingMask = 1 << 0;

	private final float[] backgroundColor = new float[4];
	private float fovY = 45f;
	private float nearClipPlane = 0.1f;
	private float farClipPlane = 20f;

	private final float[] viewMatrix = new float[16];
	private final float[] projMatrix = new float[16];
	private final float[] viewProjMatrix = new float[16];

	private final RenderingValues renderingPrms = new RenderingValues();

	Camera() {
		this.enabled = true;
		this.renderTexture = null;
		this.fbHandle = -1;
		this.depthFbHandle = -1;

		setBackgroundColor(0.5f, 0.5f, 0.5f, 1.0f);
		// Defaults to rendering to texture.
		this.postEffects.setRenderingMode(ImageEffects.FinalDestination.TEXTURE);
	}

	public void setEnabled(boolean enabling) {
		this.enabled = enabling;
	}

	public boolean isEnabled() {
		return this.enabled;
	}

	public void setRenderTexture(RenderTexture renderTexture) {
		if (this.renderTexture != renderTexture) {
			this.renderTexture = renderTexture;
			// Init methods may reset the rendering mode, so backup here for future restore.
			final ImageEffects.FinalDestination renderingMode = this.postEffects.getRenderingMode();
			setupFBO();
			updateProj();
			if (!this.postEffects.isInitialized()) {
				this.postEffects.init();
			}
			this.postEffects.setRenderTexture(renderTexture);
			// Restore rendering mode.
			this.postEffects.setRenderingMode(renderingMode);
		}
	}

	private void setupFBO() {
		// Deletes the previous FBO, if any.
		freeResources();

		final int[] names = { -1 };
		// FB
		GLES20.glGenFramebuffers(1, names, 0);
		this.fbHandle = names[0];
		// depth
		GLES20.glGenRenderbuffers(1, names, 0);
		this.depthFbHandle = names[0];
		GLES20.glBindRenderbuffer(GLES20.GL_RENDERBUFFER, this.depthFbHandle);
		GLES20.glRenderbufferStorage(
				GLES20.GL_RENDERBUFFER, GLES20.GL_DEPTH_COMPONENT16,
				this.renderTexture.getWidth(), this.renderTexture.getHeight());
		GLES20.glBindRenderbuffer(GLES20.GL_RENDERBUFFER, 0);
	}

	/**
	 * Free resources used by this camera: the frame buffer and render buffers
	 * are deleted.
	 */
	public void freeResources() {
		final int[] rendBuff = new int[] { this.fbHandle };
		if (this.fbHandle > 0) {
			GLES20.glDeleteFramebuffers(1, rendBuff, 0);
			this.fbHandle = -1;
		}
		if (this.depthFbHandle > 0) {
			rendBuff[0] = this.depthFbHandle;
			GLES20.glDeleteRenderbuffers(1, rendBuff, 0);
			this.depthFbHandle = 1;
		}
		// Same thing for post effects.
		this.postEffects.freeResources();
	}

	public RenderTexture getSharedRenderTexture() {
		return this.renderTexture;
	}

	public void setAsRenderingOnScreen() {
		if (currentCamera != null) {
			currentCamera.postEffects.setRenderingMode(ImageEffects.FinalDestination.TEXTURE);
			SceneConductor.getInstance().removeSurfaceChangedListener(currentCamera);
		}
		currentCamera = this;
		this.postEffects.setRenderingMode(ImageEffects.FinalDestination.SCREEN);
		SceneConductor.getInstance().addSurfaceChangedListener(this);
	}

	public boolean isRenderingOnScreen() {
		return this == currentCamera;
	}

	/**
	 * Sets the culling mask. The default value of the culling mask is
	 * <code>1&lt;&lt;0</code>.
	 * @param mask
	 * @see getCullingMask
	 * @see isMaskCulled
	 */
	public void setCullingMask(int mask) {
		this.cullingMask = mask;
	}

	/**
	 * Gets the culling mask.
	 * @return the value of the culling mask.
	 * @see setCullingMask
	 * @see isMaskCulled
	 */
	public int getCullingMask() {
		return this.cullingMask;
	}

	/**
	 * If the bitwise <code>AND</code> operation between this mask and a
	 * {@link SceneObject} layer returns <code>0</code>, then the
	 * {@link SceneObject} will not be rendered.
	 * @see setCullingMask
	 * @see getCullingMask
	 */
	public boolean isMaskCulled(SceneObject sob) {
		return (sob.getLayer() & this.cullingMask) == 0;
	}

	@Override
	public void onSurfaceChanged(int width, int height) {
		updateProj();
	}

	public int getWidth() {
		return this.renderTexture.getWidth();
	}

	public int getHeight() {
		return this.renderTexture.getHeight();
	}

	public void setNearAndFarClipPlanes(float near, float far) {
		this.nearClipPlane = near;
		this.farClipPlane = far;

		// It is needed to recompute the projection matrix.
		this.needsUpdateProj = true;
	}

	public float getNearClipPlane() {
		return this.nearClipPlane;
	}

	public float getFarClipPlane() {
		return this.farClipPlane;
	}

	public float[] getProjMatrix(float[] projMatrix) {
		System.arraycopy(this.projMatrix, 0, projMatrix, 0, 16);
		return projMatrix;
	}

	protected void updateProj() {
		final int width = getWidth();
		final int height = getHeight();

		this.renderingPrms.setViewSize(width, height);

		float ratio =  ((float) width) / height;
		perspectiveM(this.projMatrix, 0, this.fovY, ratio, this.nearClipPlane, this.farClipPlane);

		// Sets the projection matrix.
		this.renderingPrms.setProjMatrix(this.projMatrix);
	}

	public void setBackgroundColor(float r, float g, float b, float a) {
		this.backgroundColor[0] = r;
		this.backgroundColor[1] = g;
		this.backgroundColor[2] = b;
		this.backgroundColor[3] = a;
	}

	/**
	 * Visits all the current {@link SceneObject} objects to find which ones are
	 * visible by this {@link Camera}.
	 */
	private static final class SceneObjectVisitorForVisible implements SceneConductor.IRenderQueueVisitor {
		private final Camera camera;
		private ArrayList<SceneObject> visibleObjects;

		public SceneObjectVisitorForVisible(Camera camera) {
			this.camera = camera;
		}

		public void setResultArray(ArrayList<SceneObject> visibleObjects) {
			this.visibleObjects = visibleObjects;
		}

		@Override
		public void visit(SceneObject sob, int queue) {
			if (sob.isEnabled() && !this.camera.isMaskCulled(sob) && sob.getSharedRenderer() != null) {
				this.visibleObjects.add(sob);
			}
		}
	}

	private final SceneObjectVisitorForVisible sceneObjectVisitorForVisible =
			new SceneObjectVisitorForVisible(this);

	/**
	 * Adds to <code>result</code> all the {@link SceneObject} objects that are
	 * found to be visible by this {@link Camera}.
	 * @param result
	 * @param queueIndex
	 * @return
	 */
	public ArrayList<SceneObject> findVisibleSceneObjects(
			final ArrayList<SceneObject> result, int queueIndex) {
		this.sceneObjectVisitorForVisible.setResultArray(result);
		SceneConductor.getInstance().visitRenderQueue(
				queueIndex, this.sceneObjectVisitorForVisible);
		return result;
	}

	private static final Matrix4 modelViewRToTex = Matrix4.getNewIdentity();

	private final ArrayList<SceneObject> visibleSobsRToTex = new ArrayList<SceneObject>();

	/**
	 * Stores SceneObjects and their center in camera space coordinates
	 * so instances of this class can be ordered on axe Z (-Z towards +Z).
	 */
	private static final class OrderedOnZSob implements Comparable<OrderedOnZSob> {
		public SceneObject sob;
		public final Vector3 center = Vector3.getNewZeroVec();
		public boolean used = false;

		/**
		 * Sets the SceneObject that will be sorted and its current center.
		 * It will be needed to transform this center later to make the
		 * method <code>compare</code> useful.
		 * @param sob the SceneObject to render later.
		 * @param center the current center of the SceneObject, not transformed yet.
		 */
		public void setContent(SceneObject sob, Vector3 center) {
			this.sob = sob;
			this.center.setValues(center);
			this.used = true;
		}

		@Override
		public int compareTo(OrderedOnZSob other) {
			final float thisZ = this.center.getZ();
			final float otherZ = other.center.getZ();
			return this.used ? -1 : (thisZ < otherZ ? -1 : (thisZ > otherZ ? 1 : 0));
		}
	};
	private OrderedOnZSob[] orderedSobs = new OrderedOnZSob[0];

	public void renderToTexture() {
		preRenderToTexture();

		// Sets the eye position (in word space)
		this.renderingPrms.setEyePosition(getSharedTransform().getSharedWorldTranslation());

		// Sets the view matrix (i.e. inverse of the camera transform).
		getSharedTransform().getInverseWorldMatrix4(this.viewMatrix);
		this.renderingPrms.setViewMatrix(this.viewMatrix);

		// Proj x view
		Matrix.multiplyMM(this.viewProjMatrix, 0, this.projMatrix, 0, this.viewMatrix, 0);
		this.renderingPrms.setViewProjMatrix(this.viewProjMatrix);

		this.renderTexture.markAsNotSwapped();

		synchronized (modelViewRToTex) {
			getSharedTransform().getInverseWorldMatrix4(modelViewRToTex);

			final int nbQueues = SceneConductor.getInstance().countRenderQueues();
			for (int queueIndex = 0; queueIndex < nbQueues; queueIndex++) {
				visibleSobsRToTex.clear();
				findVisibleSceneObjects(visibleSobsRToTex, queueIndex);

				final boolean orderOnZ = (
						SceneConductor.getInstance().getQueueAtIndex(queueIndex) >
						Shader.RenderQueueType.GEOMETRY.index());

				if (!orderOnZ) {
					boolean needsPass = true;
					for (int numPass = 0; needsPass; numPass++) {
						needsPass = false;
						for (int iSob = 0, iSobEnd = visibleSobsRToTex.size(); iSob < iSobEnd; iSob++) {
							final Renderer renderer = visibleSobsRToTex.get(iSob).getSharedRenderer();
							if (numPass == 0) {
								renderer.setToFirstPass();
							}
							needsPass = renderer.renderOnePass(this) || needsPass;
						}
					}
				} else {
					// Create empty slots if needed ('orderedSobs' may need to be enlarged).
					if (visibleSobsRToTex.size() > this.orderedSobs.length) {
						final OrderedOnZSob[] newOrderedSobs = new OrderedOnZSob[visibleSobsRToTex.size()];
						// Copy current content.
						System.arraycopy(this.orderedSobs, 0, newOrderedSobs, 0, this.orderedSobs.length);
						// Creates the slot...
						for (int j = this.orderedSobs.length, jEnd = newOrderedSobs.length; j < jEnd; j++) {
							newOrderedSobs[j] = new OrderedOnZSob();
						}
						// Switch the arrays.
						this.orderedSobs = newOrderedSobs;
					}
					for (int iSob = 0, iSobEnd = visibleSobsRToTex.size(); iSob < iSobEnd; iSob++) {
						final SceneObject sob = visibleSobsRToTex.get(iSob);
						final OrderedOnZSob sobWrapper = this.orderedSobs[iSob];
						sobWrapper.setContent(
								sob,
								sob.getSharedTransform().getSharedWorldTranslation());
						modelViewRToTex.transformPoint(
								sobWrapper.center, sobWrapper.center);
					}
					// Sorts from negative to positive numbers, so the Z highest
					// values will appeared last and hence will be the last drawn.
					Arrays.sort(this.orderedSobs);
					// Finally renders.
					boolean needsPass = true;
					for (int numPass = 0; needsPass; numPass++) {
						needsPass = false;
						for (int iSobWrap = 0, iSobWrapEnd = this.orderedSobs.length; iSobWrap < iSobWrapEnd; iSobWrap++) {
							final OrderedOnZSob sobWrapper = this.orderedSobs[iSobWrap];
							if (sobWrapper.used) {
								final Renderer renderer = sobWrapper.sob.getSharedRenderer();
								if (numPass == 0) {
									renderer.setToFirstPass();
								}
								needsPass = renderer.renderOnePass(this) || needsPass;
								// Forget this wrapper;
								sobWrapper.used = false;
							}
						}
					}
				}
			}
		}

		GLES20.glBindFramebuffer(GLES20.GL_FRAMEBUFFER, 0);
		this.renderTexture.swapPrevAndCurrentTexture();
	}

	private void preRenderToTexture() {
		if (this.fbHandle < 1) {
			throw new Utils.WrongFrameBufferHandleException(
					"FBO not initialized? (fbHandle = " + this.fbHandle + ")");
		}

		GLES20.glBindFramebuffer(GLES20.GL_FRAMEBUFFER, this.fbHandle);

		final int renderTextureHandle = this.renderTexture.getTextureHandle();

		GLES20.glFramebufferTexture2D(
				GLES20.GL_FRAMEBUFFER, GLES20.GL_COLOR_ATTACHMENT0,
                GLES20.GL_TEXTURE_2D, renderTextureHandle, 0);

		GLES20.glFramebufferRenderbuffer(
				GLES20.GL_FRAMEBUFFER, GLES20.GL_DEPTH_ATTACHMENT,
				GLES20.GL_RENDERBUFFER, this.depthFbHandle);

		final int fbStatus = GLES20.glCheckFramebufferStatus(GLES20.GL_FRAMEBUFFER);
		if (fbStatus != GLES20.GL_FRAMEBUFFER_COMPLETE) {
			throw new Utils.FrameBufferIncompleteException(
					"Could not bind frame buffer: " + fbStatus);
		}

		if (this.needsUpdateProj) {
			updateProj();
			this.needsUpdateProj = false;
		}

		GLES20.glViewport(0, 0, this.renderTexture.getWidth(), this.renderTexture.getHeight());
		GLES20.glDepthRangef(this.nearClipPlane, this.farClipPlane);
		// Set the background frame color
		GLES20.glClearColor(
				this.backgroundColor[0],
				this.backgroundColor[1],
				this.backgroundColor[2],
				this.backgroundColor[3]);
		GLES20.glClearDepthf(1.0f);
    	GLES20.glClear(GLES20.GL_COLOR_BUFFER_BIT | GLES20.GL_DEPTH_BUFFER_BIT);
	}

	/**
	 * Gets parameters that can be passed to all shaders for a rendering with
	 * this {@link Camera}. Note that info contained by the returned object will
	 * be updated <b>only</b> by the method <code>render</code>.
	 *
	 * @return an object that contains info for rendering.
	 */
	public RenderingValues getSharedRenderingValues() {
		return this.renderingPrms;
	}

	/**
     * Define a projection matrix in terms of a field of view angle, an
     * aspect ratio, and z clip planes.
     * This code has been extracted from Android API level 14.
     *
     * @param m the float array that holds the perspective matrix
     * @param offset the offset into float array m where the perspective
     * matrix data is written
     * @param fovy field of view in y direction, in degrees
     * @param aspect width to height aspect ratio of the viewport
     * @param zNear
     * @param zFar
     */
    public static void perspectiveM(float[] m, int offset,
          float fovy, float aspect, float zNear, float zFar) {
        float f = 1.0f / (float) Math.tan(fovy * (Math.PI / 360.0));
        float rangeReciprocal = 1.0f / (zNear - zFar);

        m[offset + 0] = f / aspect;
        m[offset + 1] = 0.0f;
        m[offset + 2] = 0.0f;
        m[offset + 3] = 0.0f;

        m[offset + 4] = 0.0f;
        m[offset + 5] = f;
        m[offset + 6] = 0.0f;
        m[offset + 7] = 0.0f;

        m[offset + 8] = 0.0f;
        m[offset + 9] = 0.0f;
        m[offset + 10] = (zFar + zNear) * rangeReciprocal;
        m[offset + 11] = -1.0f;

        m[offset + 12] = 0.0f;
        m[offset + 13] = 0.0f;
        m[offset + 14] = 2.0f * zFar * zNear * rangeReciprocal;
        m[offset + 15] = 0.0f;
    }

    public void useAlternativePostEffectShader(Shader shader) {
    	this.postEffects.useAlternativeShader(shader);
    }

    public void setToFirstPostEffectPass() {
    	this.postEffects.setToFirstPass();
    }

    public int getCurrentPostEffectPass() {
    	return this.postEffects.getCurrentPass();
    }

    public boolean renderOnePostEffectPass() {
    	return this.postEffects.renderOnePass();
    }

    public boolean postEffectStillNeedsPass() {
    	return this.postEffects.stillNeedsPass();
    }
}
