package com.laguerre.kan;

import android.util.SparseArray;

public final class Lights {
	private static Lights instance = null;
	public static Lights getInstance() {
		if (instance == null) {
			instance = new Lights();
		}
		return instance;
	}
	public static void resetInstance() {
		instance = null;
	}

	public static final int MAX_PER_PIXEL_LIGHTS = 2;
	private static final float[] DEFAULT_ATTENUATION_PARAMS = { 0f, 0f, 0f, 0f };
	private static final float[] DEFAULT_COSINE_CONE_PARAMS = { 0f, 0f };
	private final SparseArray<SceneObject> lights = new SparseArray<SceneObject>();

	private Lights() {
	}

	public void addLight(SceneObject sob) {
		this.lights.put(sob.getId(), sob);
	}

	public void removeLight(SceneObject sob) {
		this.lights.remove(sob.getId());
	}

	public void removeLights() {
		this.lights.clear();
	}

	public int lightCount() {
		return this.lights.size();
	}

	/**
	 * Passes the following values for each known light (max: <code>MAX_PER_PIXEL_LIGHTS</code>)
	 * to the given <code>shader</code>:
	 * <ul>
	 * <li><code>lightType</code> with: 0 => Directional, 1 => Point, 2 => Spot</li>
	 * <li><code>position</code>, which is the light position in world space</li>
	 * <li><code>direction</code>, which is the negated direction of the light
	 *  in world space (used only with directional and spot lights)</li>
	 * <li><code>diffuseColor</code>, which is the diffuse color</li>
	 * <li><code>specularColor</code>, which is the specular color (used only
	 *  with point and specular lights)</li>
	 * <li><code>attenuation</code>, which is a <code>vec4</code> with the following
	 *   values and attenuation computed as 1.0/(Kc + Kl*d + Kq*d²):
	 *   <ul>
	 * 	   <li><code>range</code> (index 0), which is a value to cut the light
	 *       after this distance (used only with spot lights)</li>
	 *     <li><code>constant attenuation -- Kc</code> (index 1)</li>
	 *     <li><code>linear attenuation -- Kl</code> (index 2)</li>
	 *     <li><code>quadratic attenuation -- Kq</code> (index 3)</li>
	 *   </ul>
	 * <li><code>cosConeValues</code>, which is a <code>vec2</code> with the
	 *   following values:
	 *   <ul>
	 *     <li><code>cosInnerCone</code> (index 0), which is the cosine of the
	 *       inner cut-off angle of a spot light</li>
	 *     <li><code>cosOuterCone</code> (index 1), which is the cosine of the
	 *       outer cut-off angle of a spot light.</li>
	 *   </ul>
	 * </ul>
	 * If less than <code>MAX_PER_PIXEL_LIGHTS</code> lights have been provided,
	 * values of default lights will be passed: directional light with black colors,
	 * zero direction and zero range.
	 * If more than <code>MAX_PER_PIXEL_LIGHTS</code> lights have been provided,
	 * then only the first <code>MAX_PER_PIXEL_LIGHTS</code> lights will be used
	 * (the order is undefined).
	 *
	 * @param camera the current camera so lights from ignored layers are discarded.
	 * @param shader the shader that will receive the values.
	 * @see MAX_PER_PIXEL_LIGHTS
	 */
	public void passLightsToShader(Camera camera, Shader shader) {
		final int lightSize = this.lights.size();
		int lightNum = 0;
		for (int lightIndex = 0; lightIndex < lightSize && lightNum < MAX_PER_PIXEL_LIGHTS; lightIndex++) {
			final SceneObject sob = this.lights.valueAt(lightIndex);
			if (!camera.isMaskCulled(sob)) {
				final Light light = sob.getSharedLight();
				passLightToShader(
						shader, lightNum, light.getLightType(),
						sob.getSharedTransform().getSharedWorldTranslation(),
						light.getSharedDirection(),
						light.getSharedDiffuseColor(),
						light.getSharedSpecularColor(),
						light.getSharedAttenuationValues(),
						light.getSharedCosineAngleValues());
				lightNum++;
			}
		}
		// Puts default values if needed.
		for ( ; lightNum < MAX_PER_PIXEL_LIGHTS; lightNum++) {
			passLightToShader(
					shader, lightNum, Light.LightType.DIRECTIONAL,
					Vector3.ZERO, Vector3.ZERO,
					Color.GREEN, Color.BLACK,
					DEFAULT_ATTENUATION_PARAMS,
					DEFAULT_COSINE_CONE_PARAMS);
		}
	}

	/**
	 * Uses {@link String} to passes values associated to variables to a {@link Shader}.
	 * Internally, this method avoids to build many temporary Strings by the
	 * help of the method <code>getLightString</code> that stores/retrieves
	 * previous Strings.
	 *
	 * @param shader the shader that will receive values.
	 * @param lightNum the number of the light that is processed.
	 * @param lightType the type of the light being processed.
	 * @param position the position of the light being processed.
	 * @param direction the direction of the light being processed, if it is directional.
	 * @param diffuseColor the diffuse color of the light being processed.
	 * @param specularColor the specular color of the light being processed.
	 * @param attenuationValues array with range, constant attenuation, linear
	 *   attenuation and quadratic attenuation.
	 * @param cosineAngleValues array with the cosine inner and outer cone values.
	 */
	private void passLightToShader(
			Shader shader, int lightNum, Light.LightType lightType,
			Vector3 position, Vector3 direction,
			Color diffuseColor, Color specularColor,
			float[] attenuationValues,
			float[] cosineAngleValues) {
		// Gets strings such as
		//  "lights[0].position", "lights[0].color", etc.
		// and passes associated values.
		shader.setShaderValue(getLightString(lightNum, LightStringPrmId.LIGHT_TYPE.ordinal()), lightType.ordinal());
		shader.setShaderVector(getLightString(lightNum, LightStringPrmId.POSITION.ordinal()), position);
		shader.setShaderVector(getLightString(lightNum, LightStringPrmId.DIRECTION.ordinal()), direction);
		shader.setShaderVector(getLightString(lightNum, LightStringPrmId.DIFFUSE_COLOR.ordinal()), diffuseColor);
		shader.setShaderVector(getLightString(lightNum, LightStringPrmId.SPECULAR_COLOR.ordinal()), specularColor);
		shader.setShaderVector(getLightString(lightNum, LightStringPrmId.ATTENUATION_VALUES.ordinal()), attenuationValues);
		shader.setShaderVector(getLightString(lightNum, LightStringPrmId.COSINE_CONE_VALUES.ordinal()), cosineAngleValues);
	}

	private String getLightString(int lightNum, int paramId) {
		LightStringValues strValues = lightParamStrings.get(lightNum);
		if (strValues == null) {
			strValues = new LightStringValues(lightNum);
			lightParamStrings.append(lightNum, strValues);
		}
		return strValues.getString(paramId);
	}

	private static final SparseArray<LightStringValues> lightParamStrings = new SparseArray<LightStringValues>();

	private static enum LightStringPrmId {
		LIGHT_TYPE,
		POSITION,
		DIRECTION,
		DIFFUSE_COLOR,
		SPECULAR_COLOR,
		ATTENUATION_VALUES,
		COSINE_CONE_VALUES
	}

	private static final class LightStringValues {
		private final SparseArray<String> stringValues = new SparseArray<String>();

		LightStringValues(int lightNum) {
			final String lightsStr = "lights[" + lightNum + "].";
			this.stringValues.put(LightStringPrmId.LIGHT_TYPE.ordinal(), lightsStr + "lightType");
			this.stringValues.put(LightStringPrmId.POSITION.ordinal(), lightsStr + "position");
			this.stringValues.put(LightStringPrmId.DIRECTION.ordinal(), lightsStr + "direction");
			this.stringValues.put(LightStringPrmId.DIFFUSE_COLOR.ordinal(), lightsStr + "diffuseColor");
			this.stringValues.put(LightStringPrmId.SPECULAR_COLOR.ordinal(), lightsStr + "specularColor");
			this.stringValues.put(LightStringPrmId.ATTENUATION_VALUES.ordinal(), lightsStr + "attenuation");
			this.stringValues.put(LightStringPrmId.COSINE_CONE_VALUES.ordinal(), lightsStr + "cosConeValues");
		}

		String getString(int id) {
			return this.stringValues.get(id);
		}
	}
}
