package com.laguerre.kan;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.ShortBuffer;
import java.util.ArrayList;

public abstract class Mesh extends Component {
	public static final int POSITION_DATA_SIZE = 4;
	public static final int NORMAL_DATA_SIZE = 3;
	public static final int TANGENT_DATA_SIZE = 3;
	public static final int UV_DATA_SIZE = 2;

	protected static final class SubMeshData {
		public float[] vertices = new float[0];
		public float[] normals = new float[0];
		public float[] tangents = new float[0];
		public float[] uvs = new float[0];
		public short[] triangles = new short[0];

		public FloatBuffer packedData;
		public ShortBuffer drawListBuffer;

		public int getVertexPackedDataSize() {
			int dataSize = POSITION_DATA_SIZE;
			if (hasNormals()) {
				dataSize += NORMAL_DATA_SIZE;
			}
			if (hasTangents()) {
				dataSize += TANGENT_DATA_SIZE;
			}
			if (hasUVs()) {
				dataSize += UV_DATA_SIZE;
			}
			// Adds padding if needed to reach a multiple of 4 bytes.
			dataSize += (dataSize % 4);
			return dataSize;
		}

		public boolean hasNormals() {
			return this.normals.length > 0;
		}

		public boolean hasTangents() {
			return this.tangents.length > 0;
		}

		public boolean hasUVs() {
			return this.uvs.length > 0;
		}

		public void build() {
	    	final int vertexCount = vertices.length / POSITION_DATA_SIZE;
	    	final int dataSize = getVertexPackedDataSize();

	    	final float[] vertexData = new float[vertexCount * dataSize];
	    	for (int i = 0; i < vertexCount; i++) {
	    		int writeCursor = i * dataSize;

	    		System.arraycopy(vertices, POSITION_DATA_SIZE * i, vertexData, writeCursor, POSITION_DATA_SIZE);
	    		writeCursor += POSITION_DATA_SIZE;

	    		if (hasNormals()) {
	    			System.arraycopy(normals, NORMAL_DATA_SIZE * i, vertexData, writeCursor, NORMAL_DATA_SIZE);
		    		writeCursor += NORMAL_DATA_SIZE;
	    		}
	    		if (hasTangents()) {
	    			System.arraycopy(tangents, TANGENT_DATA_SIZE * i, vertexData, writeCursor, TANGENT_DATA_SIZE);
		    		writeCursor += TANGENT_DATA_SIZE;
	    		}
	    		if (hasUVs()) {
	    			System.arraycopy(uvs, UV_DATA_SIZE * i, vertexData, writeCursor, UV_DATA_SIZE);
		    		writeCursor += UV_DATA_SIZE;
	    		}
	    	}

	    	final ByteBuffer bb = ByteBuffer.allocateDirect(vertexCount * dataSize * Float.SIZE / 8);
	    	bb.order(ByteOrder.nativeOrder());
	    	this.packedData = bb.asFloatBuffer();
	    	this.packedData.put(vertexData);
	    	this.packedData.position(0);

	    	final ByteBuffer bbDrawList = ByteBuffer.allocateDirect(triangles.length * Short.SIZE / 8);
	    	bbDrawList.order(ByteOrder.nativeOrder());
	    	this.drawListBuffer = bbDrawList.asShortBuffer();
	    	this.drawListBuffer.put(triangles);
	    	this.drawListBuffer.position(0);
		}

		public void recomputeNormals() {
			final int vertexCount = vertices.length / POSITION_DATA_SIZE;

			if (this.normals.length / NORMAL_DATA_SIZE != vertexCount) {
				// Not enough place -- or too much -- to store all the coordinates
				// of the normals so we create a new array.
				this.normals = new float[vertexCount * NORMAL_DATA_SIZE];
			}

			final Vector3 p0 = Vector3.getNewZeroVec();
			final Vector3 p1 = Vector3.getNewZeroVec();
			final Vector3 p2 = Vector3.getNewZeroVec();

			final Vector3 v10 = Vector3.getNewZeroVec();
			final Vector3 v12 = Vector3.getNewZeroVec();
			final Vector3 normal = Vector3.getNewZeroVec();

			for (int i = 0, iEnd = this.triangles.length; i < iEnd; i += 3) {
				final int p0Index = this.triangles[i    ];
				final int p1Index = this.triangles[i + 1];
				final int p2Index = this.triangles[i + 2];

				int startVertex = POSITION_DATA_SIZE * p0Index;
				p0.setValues(
						this.vertices[startVertex    ],
						this.vertices[startVertex + 1],
						this.vertices[startVertex + 2]);
				p0.mulSelf(1.0f / this.vertices[startVertex + 3]);
				startVertex = POSITION_DATA_SIZE * p1Index;
				p1.setValues(
						this.vertices[startVertex    ],
						this.vertices[startVertex + 1],
						this.vertices[startVertex + 2]);
				p1.mulSelf(1.0f / this.vertices[startVertex + 3]);
				startVertex = POSITION_DATA_SIZE * p2Index;
				p2.setValues(
						this.vertices[startVertex    ],
						this.vertices[startVertex + 1],
						this.vertices[startVertex + 2]);
				p2.mulSelf(1.0f / this.vertices[startVertex + 3]);

				v10.setValues(p0);
				v10.minusSelf(p1);

				v12.setValues(p2);
				v12.minusSelf(p1);

				Vector3.cross(normal, v12, v10);
				normal.normalizeSelf();

				int startNormal = NORMAL_DATA_SIZE * p0Index;
				this.normals[startNormal    ] = normal.getX();
				this.normals[startNormal + 1] = normal.getY();
				this.normals[startNormal + 2] = normal.getZ();
				startNormal = NORMAL_DATA_SIZE * p1Index;
				this.normals[startNormal    ] = normal.getX();
				this.normals[startNormal + 1] = normal.getY();
				this.normals[startNormal + 2] = normal.getZ();
				startNormal = NORMAL_DATA_SIZE * p2Index;
				this.normals[startNormal    ] = normal.getX();
				this.normals[startNormal + 1] = normal.getY();
				this.normals[startNormal + 2] = normal.getZ();
			}
		}
	}

	private final ArrayList<SubMeshData> subMeshDataSet = new ArrayList<SubMeshData>();

	protected final SubMeshData getOrCreateSubMeshData(int subMeshId) {
		for (int i = this.subMeshDataSet.size(), iEnd = subMeshId + 1; i < iEnd; i++) {
			this.subMeshDataSet.add(new SubMeshData());
		}
		return this.subMeshDataSet.get(subMeshId);
	}

	protected static float[] copyToOrCreateArray(float[] srcArray, float[] dstArray) {
		if (dstArray.length != srcArray.length) {
			dstArray = srcArray.clone();
		} else {
			System.arraycopy(srcArray, 0, dstArray, 0, srcArray.length);
		}
		return dstArray;
	}

	protected static short[] copyToOrCreateArray(short[] srcArray, short[] dstArray) {
		if (dstArray.length != srcArray.length) {
			dstArray = srcArray.clone();
		} else {
			System.arraycopy(srcArray, 0, dstArray, 0, srcArray.length);
		}
		return dstArray;
	}

	public final void setVertices(float[] vertices, int subMeshId) {
		final SubMeshData meshData = getOrCreateSubMeshData(subMeshId);
		meshData.vertices = copyToOrCreateArray(vertices, meshData.vertices);
	}

	public final float[] getSharedVertices(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).vertices;
	}

	public final float[] getVertices(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).vertices.clone();
	}

	public final boolean hasNormals(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).hasNormals();
	}

	public final void setNormals(float[] normals, int subMeshId) {
		final SubMeshData meshData = getOrCreateSubMeshData(subMeshId);
		meshData.normals = copyToOrCreateArray(normals, meshData.normals);
	}

	public final float[] getSharedNormals(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).normals;
	}

	public final float[] getNormals(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).normals.clone();
	}

	public final boolean hasTangents(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).hasTangents();
	}

	public final void setTangents(float[] tangents, int subMeshId) {
		final SubMeshData meshData = getOrCreateSubMeshData(subMeshId);
		meshData.tangents = copyToOrCreateArray(tangents, meshData.tangents);
	}

	public final float[] getSharedTangents(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).tangents;
	}

	public final float[] getTangents(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).tangents.clone();
	}

	public final boolean hasUVs(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).hasUVs();
	}

	public final void setUVs(float[] uvs, int subMeshId) {
		final SubMeshData meshData = getOrCreateSubMeshData(subMeshId);
		meshData.uvs = copyToOrCreateArray(uvs, meshData.uvs);
	}

	public final float[] getSharedUVs(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).uvs;
	}

	public final float[] getUVs(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).uvs.clone();
	}

	public final void setTriangles(short[] triangles, int subMeshId) {
		final SubMeshData meshData = getOrCreateSubMeshData(subMeshId);
		meshData.triangles = copyToOrCreateArray(triangles, meshData.triangles);
	}

	public final short[] getSharedTriangles(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).triangles;
	}

	public final short[] getTriangles(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).triangles;
	}

	/**
	 * Recomputes the normals. This method can be used only once the method
	 * <code>setTriangles</code> has been called because it will use triangles.
	 * In addition, this method will take care of creating a new array to store
	 * the coordinates of the normals if the old one does not have the exact size.
	 * @param subMeshId
	 */
	public final void recomputeNormals(int subMeshId) {
		this.subMeshDataSet.get(subMeshId).recomputeNormals();
	}

	public void clear() {
		this.subMeshDataSet.clear();
	}

	public void build() {
		for (int i = 0, iEnd = this.subMeshDataSet.size(); i < iEnd; i++) {
			this.subMeshDataSet.get(i).build();
		}
	}

	public final int getNbOfSubMeshes() {
		return this.subMeshDataSet.size();
	}

	public int getVertexPackedDataSize(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).getVertexPackedDataSize();
	}

	public final int getVertexCount(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).vertices.length / POSITION_DATA_SIZE;
	}

	public final int getVertexCoordinatesCount(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).vertices.length;
	}

	public final FloatBuffer getSharedVertexPackedData(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).packedData;
	}

	public final int getDrawListCount(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).triangles.length;
	}

	public final ShortBuffer getSharedDrawListBuffer(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).drawListBuffer;
	}
}
