package com.laguerre.kan;

import java.util.Arrays;

public final class Vector3 {
	public static final Vector3 ZERO = getNewZeroVec();
	public static final Vector3 ONE = getNewOneVec();
	public static final Vector3 RIGHT = getNewRightVec();
	public static final Vector3 UP = getNewUpVec();
	public static final Vector3 FORWARD = getNewBackwardVec().negateSelf();
	public static final Vector3 BACKWARD = getNewBackwardVec();

	private final float[] mVec = new float[3];

	public Vector3() {
		this(0f, 0f, 0f);
	}

	public Vector3(float x, float y, float z) {
		setValues(x, y, z);
	}

	public Vector3(Vector3 v) {
		this(v.mVec[0], v.mVec[1], v.mVec[2]);
	}

	public void setValues(float x, float y, float z) {
		mVec[0] = x;
		mVec[1] = y;
		mVec[2] = z;
	}

	public void setValues(Vector3 v) {
		if (v == this) {
			return;
		}
		System.arraycopy(v.mVec, 0, this.mVec, 0, 3);
	}

	@Override
	public String toString() {
		return "[" + mVec[0] + ", " + mVec[1] + ", " + mVec[2] + "]";
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		} else if (!(obj instanceof Vector3)) {
			return false;
		} else {
			final Vector3 v = (Vector3) obj;
			boolean theSame = true;
			for (int i = 0; i < 3 && theSame; i++) {
				if (!Utils.fuzzyEquals(this.mVec[i], v.mVec[i])) {
					theSame = false;
				}
			}
			return theSame;
		}
	}

	@Override
	public int hashCode() {
		return Arrays.hashCode(this.mVec);
	}

	public Vector3 addSelf(float f) {
		return add(this, f, this);
	}

	public Vector3 addSelf(float dx, float dy, float dz) {
		this.mVec[0] += dx;
		this.mVec[1] += dy;
		this.mVec[2] += dz;
		return this;
	}

	public Vector3 addSelf(Vector3 v) {
		return add(this, this, v);
	}

	public Vector3 minusSelf(float dx, float dy, float dz) {
		this.mVec[0] -= dx;
		this.mVec[1] -= dy;
		this.mVec[2] -= dz;
		return this;
	}

	public Vector3 minusSelf(Vector3 v) {
		return minus(this, this, v);
	}

	public Vector3 mulSelf(float x, float y, float z) {
		this.mVec[0] *= x;
		this.mVec[1] *= y;
		this.mVec[2] *= z;
		return this;
	}

	public Vector3 mulSelf(float f) {
		return mul(this, f, this);
	}

	public Vector3 negateSelf() {
		return mul(this, -1f, this);
	}

	public Vector3 scaleSelf(Vector3 scales) {
		return scale(this, scales, this);
	}

	public Vector3 scaleSelf(float f) {
		return scaleSelf(f, f, f);
	}

	public Vector3 scaleSelf(float x, float y, float z) {
		return scale(this, x, y, z);
	}

	public Vector3 normalizeSelf() {
		return normalize(this, this);
	}

	public float dot(Vector3 v) {
		return dot(this, v);
	}

	public static Vector3 add(Vector3 result, float f, Vector3 v) {
		for (int i = 0; i < 3; i++) {
			result.mVec[i] = f + v.mVec[i];
		}
		return result;
	}

	public static Vector3 add(Vector3 result, Vector3 v0, Vector3 v1) {
		for (int i = 0; i < 3; i++) {
			result.mVec[i] = v0.mVec[i] + v1.mVec[i];
		}
		return result;
	}

	public static Vector3 minus(Vector3 result, Vector3 v0, Vector3 v1) {
		for (int i = 0; i < 3; i++) {
			result.mVec[i] = v0.mVec[i] - v1.mVec[i];
		}
		return result;
	}

	public static Vector3 mul(Vector3 result, float f, Vector3 v) {
		for (int i = 0; i < 3; i++) {
			result.mVec[i] = f * v.mVec[i];
		}
		return result;
	}

	public static Vector3 scale(Vector3 result, Vector3 scales, Vector3 v) {
		for (int i = 0; i < 3; i++) {
			result.mVec[i] = scales.mVec[i] * v.mVec[i];
		}
		return result;
	}

	public static Vector3 scale(Vector3 result, float x, float y, float z) {
		result.mVec[0] *= x;
		result.mVec[1] *= y;
		result.mVec[2] *= z;
		return result;
	}

	public static Vector3 cross(Vector3 result, Vector3 v0, Vector3 v1) {
		result.setValues(
				(v0.mVec[1] * v1.mVec[2]) - (v0.mVec[2] * v1.mVec[1]),
				(v0.mVec[2] * v1.mVec[0]) - (v0.mVec[0] * v1.mVec[2]),
				(v0.mVec[0] * v1.mVec[1]) - (v0.mVec[1] * v1.mVec[0]));
		return result;
	}

	public static Vector3 add(float f, Vector3 v) {
		final Vector3 ret = new Vector3();
		return add(ret, f, v);
	}

	public static Vector3 add(Vector3 v0, Vector3 v1) {
		final Vector3 ret = new Vector3();
		return add(ret, v0, v1);
	}

	public static Vector3 minus(Vector3 v0, Vector3 v1) {
		final Vector3 ret = new Vector3();
		return minus(ret, v0, v1);
	}

	public static Vector3 mul(float f, Vector3 v) {
		final Vector3 ret = new Vector3();
		return mul(ret, f, v);
	}

	public static Vector3 negate(Vector3 result, Vector3 v) {
		return mul(result, -1f, v);
	}

	public static Vector3 negate(Vector3 v) {
		return mul(-1f, v);
	}

	public static Vector3 scale(Vector3 scales, Vector3 v) {
		Vector3 ret = new Vector3();
		return scale(ret, scales, v);
	}

	public static float dot(Vector3 v0, Vector3 v1) {
		float ret = 0f;
		for (int i = 0; i < 3; i++) {
			ret += v0.mVec[i] * v1.mVec[i];
		}
		return ret;
	}

	public static Vector3 cross(Vector3 v0, Vector3 v1) {
		final Vector3 ret = new Vector3();
		return cross(ret, v0, v1);
	}

	public float getSquareNorm() {
		return dot(this, this);
	}

	public float getNorm() {
		return (float) Math.sqrt(getSquareNorm());
	}

	public static Vector3 normalize(Vector3 result, Vector3 v) {
		final float invNorm = 1.0f / v.getNorm();
		return mul(result, invNorm, v);
	}

	public static Vector3 normalize(Vector3 v) {
		final Vector3 ret = new Vector3();
		return normalize(ret, v);
	}

	public float getX() {
		return this.mVec[0];
	}

	public float getY() {
		return this.mVec[1];
	}

	public float getZ() {
		return this.mVec[2];
	}

	/**
	 * Gets the value of a component of this {@link Vector3}.
	 * <ul>
	 *   <li>index 0 equals to getX()</li>
	 *   <li>index 1 equals to getY()</li>
	 *   <li>index 2 equals to getZ()</li>
	 * </ul>
	 * @param i between 0 and 3 (excluded) to get x, y or z.
	 * @return the value of x, y, z.
	 */
	public float getComponent(int i) {
		return this.mVec[i];
	}

	public void getComponents(float[] results) {
		System.arraycopy(this.mVec, 0, results, 0, 3);
	}

	public static Vector3 getNewZeroVec() {
		return new Vector3(0.0f, 0.0f, 0.0f);
	}

	public static Vector3 getNewOneVec() {
		return new Vector3(1.0f, 1.0f, 1.0f);
	}

	public static Vector3 getNewRightVec() {
		return new Vector3(1.0f, 0.0f, 0.0f);
	}

	public static Vector3 getNewUpVec() {
		return new Vector3(0.0f, 1.0f, 0.0f);
	}

	public static Vector3 getNewBackwardVec() {
		return new Vector3(0.0f, 0.0f, 1.0f);
	}
}
